<?php

/**
 *  This file is part of devMX TeamSpeak3 Webviewer for Wordpress.
 *  Copyright (C) 2012  Maximilian Narr
 *
 *  devMX TeamSpeak3 Webviewer for Wordpress is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TeamSpeak3 Webviewer is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with devMX TeamSpeak3 Webviewer for Wordpress.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace devmx\Wordpress\TSViewer;

class TSViewer
{

    const VERSION = "1.3";
    const VERSION_OPTION_NAME = "tswebviewer_version";
    const TABLE_NAME = "tswebviewer";
    const MAIN_MENU_NAME = "tswv-menu";
    const CONFIG_SUB_MENU_NAME = "tswv-menu-config";
    const OPTIONS_SUB_MENU_PAGE = "tswv-menu-options";
    const FORM_NAME = "tswv-edit";
    const FORM_HIDDEN_FIELD_NAME = "tswv-hidden";
    const TEXT_DOMAIN = "tswv";

    public $pluginPath;
    public $pluginUrl;
    public $cachePath;
    public $cacheUrl;
    public $adminUrl;
    public $wpdb;
    protected $tableName;

    public function __construct($pluginPath, $pluginUrl, $wpdb)
    {
        $this->wpdb = $wpdb;
        $this->pluginPath = $pluginPath . "/";
        $this->pluginUrl = $pluginUrl;
        $this->cachePath = ABSPATH . 'wp-content/uploads/tswebviewer';
        $this->cacheUrl = content_url() . '/uploads/tswebviewer/';
        $this->adminUrl = admin_url();
        $this->tableName = $this->wpdb->prefix . self::TABLE_NAME;
        $this->registerHooks();
        $this->registerShortcodes();
    }

    /**
     * Registers hooks
     * @author Maximilian Narr
     * @since 1.0
     */
    private function registerHooks()
    {
        register_activation_hook($this->pluginPath . "TSViewer.php", array(&$this, 'install'));
        register_deactivation_hook($this->pluginPath . "TSViewer.php", array(&$this, 'uninstall'));
        add_action('init', array($this, 'setL10n'));
        add_action('admin_menu', array(&$this, 'registerMenus'));
        add_action('admin_notices', array(&$this, 'handleCacheDir'));
        add_action('plugins_loaded', array(&$this, 'upgrade'));

        add_action('widgets_init', function()
                {
                    return register_widget('devmx\Wordpress\TSViewer\TSViewerWidget');
                });
    }

    private function registerShortcodes()
    {
        add_shortcode('tswv', array(&$this, 'renderShortcode'));
        add_shortcode('tswebviewer', array(&$this, 'renderShortcode'));
    }

    /**
     * Registers admin menus
     * @author Maximilian Narr
     * @since 1.0 
     */
    public function registerMenus()
    {
        add_menu_page(__('TeamSpeak3 Webviewer', 'tswv'), __('TS3 Webviewer', 'tswv'), 'manage_options', self::MAIN_MENU_NAME, array(&$this, 'mainMenu'), $this->pluginUrl . 'img/admin-menu.png');

        add_submenu_page(self::MAIN_MENU_NAME, __('Add configuration', 'tswv'), __('Add configuration', 'tswv'), 'manage_options', self::CONFIG_SUB_MENU_NAME, array(&$this, 'configMenu'));
    }

    /**
     * Outputs the main menu which lists all config files
     * @author Maximilian Narr
     * @since 1.0 
     */
    public function mainMenu()
    {
        $configDeleted = false;

        // Check if configs need to be deleted
        if (isset($_POST['delete']))
        {
            foreach ($_POST['delete'] as $configId)
            {
                $this->deleteConfiguration($configId);
                $configDeleted = true;
            }
        }

        $configs = $this->getConfigurations();
        $adminURL = $this->adminUrl;
        $menuName = self::CONFIG_SUB_MENU_NAME;

        require_once $this->pluginPath . "tpl/adminMenu.php";
    }

    /**
     * Outputs the edit menu
     * @author Maximilian Narr
     * @since 1.0 
     */
    public function configMenu()
    {
        if (isset($_GET['id'])) $configId = $_GET['id'];


        // Standard config
        $config = array("id" => "",
            "configname" => "",
            "serveradress" => "",
            "serverport" => "",
            "queryport" => "",
            "username" => "",
            "password" => "",
            "downloadIcons" => 1,
            "countryIcons" => 0,
            "showIcons" => 1,
            "style" => "base",
            "enableCaching" => 1,
            "cachingTime" => 180,
            "enableBlacklist" => 0,
            "enableWhitelist" => 0,
            "blacklistIds" => "",
            "whitelistIds" => "",
            "connectLinkTarget" => "");

        $configAdded = false;
        $configUpdated = false;

        // If a configuration should be updated or added
        if (isset($_POST[self::FORM_HIDDEN_FIELD_NAME]) && $_POST[self::FORM_HIDDEN_FIELD_NAME] == "Y")
        {
            $configValues = $config;

            foreach ($configValues as $key => $value)
            {
                if (isset($_POST[$key])) $configValues[$key] = $_POST[$key];
            }

            // If a configfile should be updated
            if (isset($_POST['id']) && $_POST['id'] != "")
            {
                $this->updateConfiguration($configValues);
                $configId = $_POST['id'];
                $configUpdated = true;
            }
            // If a configfile should be added
            else
            {
                $configId = $this->addConfiguration($configValues);
                $configAdded = true;
            }
        }

        // If a configfile should be edited or a new one should be added
        if (isset($configId))
        {
            $id = (int) $configId;
            $data = $this->getConfiguration($id);

            foreach ($data[0] as $key => $value) $config[$key] = $value;
        }

        $formName = self::FORM_NAME;
        $hiddenFieldName = self::FORM_HIDDEN_FIELD_NAME;

        $styles = $this->getStyles();

        require_once $this->pluginPath . "tpl/configMenu.php";
    }

    /**
     * Renders the shortcodes [tswv] and [tswebviewer]
     * @param array $atts Attributes of the shortcode
     * @return string Formatted HTML for output
     * @since 1.0
     * @author Maximilian Narr
     */
    public function renderShortcode($atts)
    {
        if (isset($atts['id']))
        {
            $id = (int) $atts['id'];

            $configuration = $this->getConfiguration($id);

            try
            {
                if (!count($configuration) == 0) return $this->renderViewer($configuration[0]);
            }
            catch (Exception $ex)
            {
                // Return error message, if an error occured
                $errorMessage = sprintf("<strong>Fatal Error:</strong> %s<br><strong>line: </strong>%s<br><strong>File: </strong>%s<br><strong>Trace: </strong><pre>%s</pre>", $ex->getMessage(), $ex->getLine(), $ex->getFile(), $ex->getTraceAsString());
                return sprintf('<div class="tswv-error">%s</div>', $errorMessage);
            }
        }
    }

    /**
     * Renders the webviewer for the shortcodes
     * @param array $configuration Configuration to render the server 
     * @return string Formatted HTML for output
     * @since 1.0
     * @author Maximilian Narr
     */
    private function renderViewer($configuration)
    {
        $c = new \devmx\TSWebViewer\RenderOptions();

        $c['root.server'] = $this->pluginPath;
        $c['root.public'] = $this->pluginUrl;

        $c['cache.file.url'] = $this->cacheUrl;
        
        $c['html.generate_head_tags'] = false;
        $c['render.query_clients'] = false;

        $c['connectlink.show'] = true;

        $c['cache.file.path'] = $this->cachePath;

        $c['style'] = "base";

        $c['ts3']['host'] = $configuration['serveradress'];

        $c['ts3']['query.port'] = $configuration['queryport'];

        $c['ts3']['vserver.port'] = $configuration['serverport'];

        // Check if username is necessary
        if (!empty($configuration['username']) && !empty($configuration['password']))
        {
            $c['ts3']['login.name'] = $configuration['username'];

            $c['ts3']['login.pass'] = $configuration['password'];
        }
        else
        {
            $c['ts3']['login.name'] = '';

            $c['ts3']['login.pass'] = '';
        }


        // Check if custom icons should be downloaded
        if ($configuration['downloadIcons'] == "1")
        {
            $c['download_custom_icons'] = true;
        }
        else
        {
            $c['download_custom_icons'] = false;
        }


        // Check for country-icons
        if ($configuration['countryIcons'] == "1")
        {
            $c['country_icons.show'] = true;
        }
        else
        {
            $c['country_icons.show'] = false;
        }


        // Check if icon should be shown
        if ($configuration['showIcons'] == "1")
        {
            $c['images.show'] = true;
        }
        else
        {
            $c['images.show'] = false;
        }


        // Style
        if (!empty($configuration['style']))
        {
            $c['style'] = $configuration['style'];
        }
        else
        {
            $c['style'] = "base";
        }


        // Check if caching should be enabled
        if ($configuration['enableCaching'] == "1")
        {
            $c['enable_caching'] = true;
        }
        else
        {
            $c['enable_caching'] = false;
        }


        // Caching time
        if (!empty($configuration['cachingTime']))
        {
            
        }
        else
        {
            
        }

        $c['render.query_clients'] = false;

        $c['filter.whitelist'] = array();


        // Blacklist
        if ($configuration['enableBlacklist'] == "1")
        {
            $c['filter.blacklist.enabled'] = true;
        }
        else
        {
            $c['filter.blacklist.enabled'] = false;
        }


        // Blacklist channels
        if (!empty($configuration['blacklistIds']))
        {
            $Ids = explode(",", $configuration['blacklistIds']);
            $c['filter.blacklist'] = $Ids;
        }
        else
        {
            $c['filter.blacklist'] = array();
        }


        // Whitelist
        if ($configuration['enableWhitelist'] == "1")
        {
            $c['filter.whitelist.enabled'] = true;
        }
        else
        {
            $c['filter.whitelist.enabled'] = false;
        }


        // Whitelist channels
        if (!empty($configuration['whitelistIds']))
        {
            $Ids = explode(",", $configuration['whitelistIds']);
            $c['filter.whitelist'] = $Ids;
        }
        else
        {
            $c['filter.whitelist'] = array();
        }

        
        // Connectlink target
        if(!empty($configuration['connectLinkTarget']))
        {
            $c['connectlink.target'] = $configuration['connectLinkTarget'];
        }

        try
        {
            return $c['renderer']->renderServer();
        }
        catch (\Exception $ex)
        {
            $errorHTML = sprintf("<div style='tswv-error'><p style='color:red; font-weight:bold;'>devMX TeamSpeak3 Webviewer Fatal Error</p><p>%s</p><strong>line: </strong>%s<br><strong>File: </strong><p>%s</p>", $ex->getMessage(), $ex->getLine(), $ex->getFile());
            $errorHTML .= '<p><a href="http://support.devmx.de" target="_blank">Click here for help &raquo;</a></p>';
            $errorHTML .= '</div>';
            return $errorHTML;
        }
    }

    /**
     * Checks if the caching path exists and if it is writable. 
     * Tries to create it if it does not exist yet.
     * Displays Errors if the directory cannot be created or is not writable
     * @since 1.0
     * @author Maximilian Narr 
     */
    public function handleCacheDir()
    {
        $cachePath = $this->cachePath;
        if (is_dir($this->cachePath))
        {
            if (!is_writable($this->cachePath)) require_once $this->pluginPath . "tpl/writeError.php";
        }

        $result = wp_mkdir_p($this->cachePath);

        if (!$result) require_once $this->pluginPath . "tpl/dirError.php";
    }

    /**
     * Loads the language files
     * @since 1.0
     * @author Maximilian Narr 
     */
    public function setL10n()
    {
        $path = basename($this->pluginPath) . "/l10n";
        load_plugin_textdomain(self::TEXT_DOMAIN, false, $path);
    }

    /**
     * Inserts the database for the TeamSpeak3 Webviewer Data
     * @since 1.0
     * @author Maximilian Narr 
     */
    public function install()
    {
        $wpdb = $this->wpdb;

        $sql = "CREATE TABLE " . $this->tableName . " (
            id INT(5) NOT NULL AUTO_INCREMENT,
            configname VARCHAR(60) NOT NULL,
            serveradress VARCHAR(40) NOT NULL,
            serverport INT(6) NOT NULL,
            queryport INT(6) NOT NULL,
            username VARCHAR(30) NOT NULL,
            password VARCHAR(60) NOT NULL,
            PRIMARY KEY  (id)
            );";

        require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

        dbDelta($sql);
        add_option(self::VERSION_OPTION_NAME, "1.0");
        $this->upgrade();
    }

    /**
     * Deletes the version option 
     * @since 1.1.2
     * @author Maximilian Narr
     */
    public function uninstall()
    {
        delete_option(self::VERSION_OPTION_NAME);
    }

    /**
     * Upgrades the table structure to the latest version
     * @author Maximilian Narr
     * @since 1.0 
     */
    public function upgrade()
    {
        $wpdb = $this->wpdb;

        // Upgrade to database version 1.3
        if (version_compare(get_option(self::VERSION_OPTION_NAME), "1.3.0.2", "<"))
        {
            $sql = "CREATE TABLE " . $this->tableName . " (
            id INT(5) NOT NULL AUTO_INCREMENT,
            configname VARCHAR(60) NOT NULL,
            serveradress VARCHAR(40) NOT NULL,
            serverport INT(6) NOT NULL,
            queryport INT(6) NOT NULL,
            username VARCHAR(30) NOT NULL,
            password VARCHAR(60) NOT NULL,
            downloadIcons BOOLEAN NOT NULL DEFAULT 1,
            countryIcons BOOLEAN NOT NULL DEFAULT 0,
            showIcons BOOLEAN NOT NULL DEFAULT 1,
            style VARCHAR(100) NOT NULL DEFAULT 'base',
            enableCaching BOOLEAN NOT NULL DEFAULT 1,
            cachingTime INT(6) NOT NULL DEFAULT 180,
            enableBlacklist BOOLEAN NOT NULL DEFAULT 0,
            enableWhitelist BOOLEAN NOT NULL DEFAULT 0,
            blacklistIds VARCHAR(200) NOT NULL,
            whitelistIds VARCHAR(200) NOT NULL,
            connectLinkTarget VARCHAR(100) NOT NULL,
            PRIMARY KEY  (id)
            );";

            require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

            dbDelta($sql);

            update_option(self::VERSION_OPTION_NAME, "1.3.0.2");
        }
    }

    /**
     * Returns an array containing all available webviewer configuratons
     * @return array All available webviewer configurations
     * @author Maximilian Narr
     * @since 1.0
     */
    private function getConfigurations()
    {
        $sql = "SELECT * FROM " . $this->tableName;

        return $this->wpdb->get_results($sql, ARRAY_A);
    }

    /**
     * Returns the requested configuration with id $id
     * @param int $id Id of the configuration
     * @return array Options of configuration with id $id 
     * @author Maximilian Narr
     * @since 1.0
     */
    private function getConfiguration($id)
    {
        $sql = "SELECT * FROM " . $this->tableName . " WHERE id = %d";
        $sql = $this->wpdb->prepare($sql, $id);
        return $this->wpdb->get_results($sql, ARRAY_A);
    }

    /**
     * Adds a new configuration to the databse
     * @param array $configValues array of configuration values
     * @return int Id of the inserted row
     * @author Maximilian Narr
     * @since 1.0
     */
    private function addConfiguration($configValues)
    {
        extract($configValues, EXTR_SKIP);
        $sql = "INSERT INTO " . $this->tableName . " (serveradress,
            serverport, 
            queryport, 
            username, 
            password, 
            configname, 
            downloadIcons, 
            countryIcons,
            showIcons,
            style,
            enableCaching, 
            cachingTime,
            enableBlacklist,
            enableWhitelist,
            blacklistIds,
            whitelistIds,
            connectLinkTarget) 
            VALUES (%s, %d, %d, %s, %s, %s, %d, %d, %d, %s, %d, %d, %d, %d, %s, %s, %s);";
        $sql = $this->wpdb->prepare($sql, $serveradress, $serverport, $queryport, $username, $password, $configname, $downloadIcons, $countryIcons, $showIcons, $style, $enableCaching, $cachingTime, $enableBlacklist, $enableWhitelist, $blacklistIds, $whitelistIds, $connectLinkTarget);
        $this->wpdb->query($sql);
        return $this->wpdb->insert_id;
    }

    /**
     * Updates a configuration
     * @param array $configValues array of configuration values
     * @return int
     * @author Maximilian Narr
     * @since 1.0
     */
    private function updateConfiguration($configValues)
    {
        extract($configValues, EXTR_SKIP);
        $sql = "UPDATE " . $this->tableName . " SET serveradress = %s, 
            serverport = %d, 
            queryport = %d, 
            username = %s, 
            password = %s, 
            configname = %s, 
            downloadIcons = %d, 
            countryIcons = %d,
            showIcons = %d,
            style = %s,
            enableCaching = %d, 
            cachingTime = %d,
            enableBlacklist = %d,
            enableWhitelist = %d,
            blacklistIds = %s,
            whitelistIds = %s,
            connectLinkTarget = %s WHERE id = %d;";
        $sql = $this->wpdb->prepare($sql, $serveradress, $serverport, $queryport, $username, $password, $configname, $downloadIcons, $countryIcons, $showIcons, $style, $enableCaching, $cachingTime, $enableBlacklist, $enableWhitelist, $blacklistIds, $whitelistIds, $connectLinkTarget, $id);
        return $this->wpdb->query($sql);
    }

    /**
     * Deleted the configuration with the id $id from the databse
     * @param int $id Id of the configuration to delete
     * @return int
     * @author Maximilian Narr
     * @since 1.0
     */
    private function deleteConfiguration($id)
    {
        $sql = "DELETE FROM " . $this->tableName . " WHERE id = %d";
        $sql = $this->wpdb->prepare($sql, $id);
        return $this->wpdb->query($sql);
    }

    /**
     * Returns an array of available styles
     * @return array List of availabel styles
     * @author Maximilian Narr
     * @since 1.3
     */
    private function getStyles()
    {
        $styles = array();
        $stylesDirectory = $this->pluginPath . 'styles';

        $files = scandir($stylesDirectory);

        foreach ($files as $file)
        {
            if ($file != "." && $file != "..")
            {
                $styles[] = $file;
            }
        }

        return $styles;
    }

}

?>
